import torch
import gym
from gym import spaces

import numpy as np

from stable_baselines3.common.on_policy_algorithm import OnPolicyAlgorithm
from stable_baselines3.common.env_util import unwrap_wrapper
from stable_baselines3.common.utils import obs_as_tensor

from ntn_torch.rl.common.policies import NTNPolicy
from ntn_torch.rl.common.buffers import EpisodicBuffer

class NTNReinforce(OnPolicyAlgorithm):
  def __init__(
    self,
    env,
    learning_rate=3e-4,
    n_epochs=1,
    gamma=0.99,
    max_grad_norm=0.5,
    tensorboard_log=None,
    create_eval_env=False,
    policy_kwargs=None,
    verbose=0,
    seed=None,
    device='cpu',
    _init_setup_model=True
  ):
    super(NTNReinforce, self).__init__(
      NTNPolicy,
      env,
      learning_rate,
      unwrap_wrapper(env.envs[0], gym.wrappers.TimeLimit)._max_episode_steps,
      gamma,
      1.0,
      1.0, # ent_coef
      1.0, # vf_coef
      max_grad_norm,
      False, # use_sde
      0, # sde_sample_freq
      policy_base=NTNPolicy,
      tensorboard_log=tensorboard_log,
      create_eval_env=create_eval_env,
      policy_kwargs=policy_kwargs,
      verbose=verbose,
      seed=seed,
      device=device,
      _init_setup_model=_init_setup_model,
      supported_action_spaces=(
        spaces.Box,
        spaces.Discrete
      )
    )

    self.episodic_buffer = None

    if _init_setup_model:
      self._setup_model()

  def _setup_model(self):
    self._setup_lr_schedule()
    self.set_random_seed(self.seed)

    self.episodic_buffer = EpisodicBuffer(self.env, gamma=self.gamma)
    self.rollout_buffer = self.episodic_buffer
    self.policy = NTNPolicy(self.env.observation_space, self.env.action_space, self.lr_schedule, **self.policy_kwargs)
    
  def collect_episode(self):
    assert self._last_obs is not None, "No previous observation was provided"
    self.policy.set_training_mode(False)
    self.episodic_buffer.reset()

    n_envs = self.env.num_envs
    dones = np.zeros(n_envs, dtype=np.bool)
    while not dones.all():
      with torch.no_grad():
        obs_tensor = obs_as_tensor(self._last_obs, self.device)
        actions, log_prob = self.policy(obs_tensor)

      # Rescale and perform action
      clipped_actions = actions.numpy()
      # Clip the actions to avoid out of bound error
      if isinstance(self.action_space, gym.spaces.Box):
          clipped_actions = np.clip(actions, self.action_space.low, self.action_space.high)

      obs, rew, done, infos = self.env.step(np.atleast_1d(clipped_actions))
      dones += done
      
      self.episodic_buffer.add(self._last_obs, clipped_actions.reshape((n_envs,1)), rew, dones, log_prob)

      self.num_timesteps += self.env.num_envs - dones.sum()
      self._last_obs = obs
      self._last_episode_starts = done
      self._update_info_buffer(infos)

    self.num_timesteps += self.env.num_envs # Accounting for the last step in each env
    self.episodic_buffer.compute_returns()

    return True

  def collect_rollouts(self, env, callback, rollout_buffer, n_rollout_steps):
    """Just passes through to collect_episode"""
    return self.collect_episode()

  def _loss(self, timestep, observation, action, ret):
    return -self.gamma**timestep*ret*self.policy.get_distribution(
        torch.atleast_2d(obs_as_tensor(observation, self.device))).log_prob(
            obs_as_tensor(action, self.device))

  def train(self):
     # Switch to train mode
    self.policy.set_training_mode(True)
    # Update optimizer learning rate
    self._update_learning_rate(self.policy.optimizer)

    for timestep, observation, action, reward, ret, log_prob in self.episodic_buffer.all():
      loss = self._loss(timestep, observation, action, ret)

      # Optimization step
      self.policy.optimizer.zero_grad()
      loss.backward()
      self.policy.optimizer.step()


import torch

from ntn_torch.net import FunctionalDiscriminator
from ntn_torch.rl.common.extractors import ThermometerExtractor

from stable_baselines3.common.policies import BasePolicy
from stable_baselines3.common.preprocessing import get_action_dim
from stable_baselines3.common.distributions import (
    CategoricalDistribution,
    DiagGaussianDistribution,
    make_proba_distribution,
)

class NTNPolicy(BasePolicy):
  def __init__(
    self,
    observation_space,
    action_space,
    lr_schedule,
    tuple_size,
    squash_output = False,
    features_extractor_class = ThermometerExtractor,
    features_extractor_kwargs = {'low': -1., 'high': 1., 'res': 16},
    optimizer_class = torch.optim.SGD,
    optimizer_kwargs = None
  ):
    super(NTNPolicy, self).__init__(
      observation_space,
      action_space,
      features_extractor_class,
      features_extractor_kwargs,
      optimizer_class=optimizer_class,
      optimizer_kwargs=optimizer_kwargs,
      squash_output=squash_output,
    )

    self.tuple_size = tuple_size
    # print("Here?")
    self.features_extractor = features_extractor_class(
        self.observation_space, **self.features_extractor_kwargs)
    # print("There??")
    self.features_dim = self.features_extractor.features_dim
    # print(self.features_dim)

    self.action_dist = make_proba_distribution(
        action_space, use_sde=False, dist_kwargs=None)

    self._build(lr_schedule)

  def _get_constructor_parameters(self):
    raise NotImplementedError()

  def _build(self, lr_schedule):
    if isinstance(self.action_dist, DiagGaussianDistribution):
      output_dim = 2*get_action_dim(self.action_space)
    elif isinstance(self.action_dist, CategoricalDistribution):
      output_dim = self.action_space.n
    else:
      raise NotImplementedError(f"Unsupported distribution '{self.action_dist}'.")

    self.policy_net = FunctionalDiscriminator(
        self.features_dim, output_dim, self.tuple_size)
    self.optimizer = self.optimizer_class(
        self.policy_net.parameters(), lr=lr_schedule(1), **self.optimizer_kwargs)
    
    # TODO: Consider having the std not be action-dependent in the gaussian case,
    # as in the library. Something like:
    # self.log_std = nn.Parameter(th.ones(self.action_dim) * log_std_init, requires_grad=True)
  def _get_action_dist_from_net_out(self, net_output):
    net_output = torch.atleast_2d(net_output)
    if isinstance(self.action_dist, DiagGaussianDistribution):
      # First half of the output is the mean vector and the second half, the
      # diagonal of the covariance matrix
      return self.action_dist.proba_distribution(*torch.tensor_split(net_output, 2))
    elif isinstance(self.action_dist, CategoricalDistribution):
      return self.action_dist.proba_distribution(action_logits=net_output)
    else:
      raise ValueError("Invalid action distribution")

  def get_distribution(self, obs):
    # obs = self.obs_to_tensor(obs)
    features = self.extract_features(obs)
    net_out  = self.policy_net(features)
    return self._get_action_dist_from_net_out(net_out)

  def forward(self, obs,  deterministic=False):
    distribution = self.get_distribution(obs)
    actions = distribution.get_actions(deterministic=deterministic)
    log_prob = distribution.log_prob(actions)
    return actions, log_prob

  def _predict(self, observation, deterministic = False):
    # observation = torch.atleast_2d(observation)
    return self.get_distribution(observation).get_actions(deterministic=deterministic)



import numpy as np
import gym

from scipy.signal import lfilter
from gym import spaces

from stable_baselines3.common.buffers import BaseBuffer
from stable_baselines3.common.vec_env import VecEnv
from stable_baselines3.common.env_util import unwrap_wrapper

class EpisodicBuffer(BaseBuffer):
  def __init__(
      self,
      env,
      gamma=0.99
  ):
    if not isinstance(env, VecEnv):
      raise ValueError("environment must be instance of VecEnv")
    
    max_steps = unwrap_wrapper(env.envs[0], gym.wrappers.TimeLimit)._max_episode_steps

    if max_steps is None:
      raise ValueError("environment must have a time limit")

    super(EpisodicBuffer, self).__init__(
        max_steps,
        env.observation_space,
        env.action_space,
        "cpu",
        env.num_envs
    )
    self.gamma = gamma
    self.max_steps = max_steps

    self.observations, self.actions, self.rewards  = None, None, None
    self.returns, self.log_probs, self.episode_end, self.env_has_ended = None, None, None, None

    self.reset()

  def reset(self):
    self.episode_end = np.full(self.n_envs, np.iinfo(np.int).max) # Points to where the episode actually ended for each env
    self.env_has_ended = np.zeros(self.n_envs, dtype=np.bool)
    self.observations = np.zeros((self.buffer_size, self.n_envs) + self.obs_shape, dtype=np.float32)
    self.actions = np.zeros((self.buffer_size, self.n_envs, self.action_dim), dtype=np.float32)
    self.rewards = np.zeros((self.buffer_size, self.n_envs), dtype=np.float32)
    self.returns = np.zeros((self.buffer_size, self.n_envs), dtype=np.float32)
    self.log_probs = np.zeros((self.buffer_size, self.n_envs), dtype=np.float32)

    super(EpisodicBuffer, self).reset()

  def add(self, obs, action, reward, dones, log_prob):
    if len(log_prob.shape) == 0:
      # Reshape 0-d tensor to avoid error
      log_prob = log_prob.reshape(-1, 1)

    # Reshape needed when using multiple envs with discrete observations
    # as numpy cannot broadcast (n_discrete,) to (n_discrete, 1)
    if isinstance(self.observation_space, spaces.Discrete):
      obs = obs.reshape((self.n_envs,) + self.obs_shape)

    self.env_has_ended += dones
    self.observations[self.pos] = np.array(obs).copy()
    self.actions[self.pos] = np.array(action).copy()
    self.rewards[self.pos] = np.array(reward).copy()
    self.log_probs[self.pos] = log_prob.clone().cpu().numpy()
    self.pos += 1
    if self.pos == self.buffer_size or self.env_has_ended.all():
        self.full = True

    self.episode_end = np.where(self.env_has_ended, np.minimum(self.episode_end, self.pos), self.episode_end)

  def compute_returns(self):
    effective_rewards = np.where(
      np.indices(self.rewards.shape)[0] >= self.episode_end.reshape(1, -1),
      0.,
      self.rewards
    )

    self.returns = lfilter([1], [1, -self.gamma], x=effective_rewards[::-1, :], axis=0)[::-1, :]

  def _get_samples(self):
    raise NotImplementedError()

  def all(self):
    """Iterate over the episodes, one env at a time and in order"""
    "timestep, observation, action, reward, return, log_prob"
    for episode in range(self.n_envs):
      tail = self.episode_end[episode] if self.env_has_ended[episode] else self.pos
      for timestep in range(tail):
        yield (
          timestep,
          self.observations[timestep, episode],
          self.actions[timestep, episode],
          self.rewards[timestep, episode],
          self.returns[timestep, episode],
          self.log_probs[timestep, episode]
        )


import numpy as np
import torch


class Thermometer(object):
    def __init__(self, minimum, maximum, resolution):
        self.minimum = minimum
        self.maximum = maximum
        self.resolution = resolution

    def __repr__(self):
        return f"Thermometer(minimum={self.minimum}, maximum={self.maximum}, resolution={self.resolution})"

    def encode(self, X):
        X = torch.as_tensor(X, device='cpu')

        enc = torch.arange(self.resolution).expand(*X.shape, self.resolution)
        threshold = (X.reshape((*X.shape, 1)) - self.minimum)/(self.maximum - self.minimum)*self.resolution

        return enc < threshold

    # def decode(self, pattern):
    #     pattern = np.asarray(pattern)

    #     # TODO: Check if pattern is at least a vector
    #     # TODO: Check if pattern length or number of rows is equal to resolution
    #     # TODO: Check if pattern is a binary array
    #     if pattern.ndim == 1:
    #         # TODO: Test np.count_nonzero
    #         popcount = np.sum(pattern)

    #         return self.minimum + popcount*(self.maximum - self.minimum)/self.resolution

    #     return np.asarray([self.decode(pattern[..., i]) for i in range(pattern.shape[-1])])


# class CircularThermometerEncoder(object):
#     def __init__(self, minimum, maximum, resolution, wrap=True):
#         self.minimum = minimum
#         self.maximum = maximum
#         self.resolution = resolution
#         self.block_len = np.floor(self.resolution/2)
#         self.wrap = wrap
#         self.max_shift = resolution if wrap else resolution - self.block_len

#     def __repr__(self):
#         return f"CircularThermometerEncoder(minimum={self.minimum}, maximum={self.maximum}, resolution={self.resolution}), wrap={self.wrap}"

#     def encode(self, X):
#         X = np.asarray(X)

#         if X.ndim == 0:
#             if X < self.minimum or X > self.maximum:
#                 raise ValueError(
#                     f"Encoded values should be in the range [{self.minimum}, {self.maximum}]. Value given: {X}")

#             base_pattern = np.fromfunction(
#                 lambda i: i < self.block_len, (self.resolution,)).astype(np.uint8)
#             shift = int(np.abs(self.minimum-X) /
#                         (self.maximum-self.minimum)*self.max_shift)

#             return np.roll(base_pattern, shift)

#         return np.stack([self.encode(v) for v in X], axis=X.ndim)

    # def decode(self, pattern):
    #     pattern = np.asarray(pattern)

    #     # TODO: Check if pattern is at least a vector
    #     # TODO: Check if pattern length or number of rows is equal to resolution
    #     # TODO: Check if pattern is a binary array
    #     if pattern.ndim == 1:
    #         first_0 = index(pattern, 0)[0]
    #         first_1 = index(pattern, 1)[0]

    #         if first_0 > first_1:
    #             shift = (first_0 - self.block_len) % self.resolution
    #         else:
    #             shift = first_1

    #         if shift > self.max_shift:
    #             raise ValueError(
    #                 "Input pattern wraps around. Consider using a encoder with wrap enabled")

    #         return self.minimum + shift*(self.maximum - self.minimum)/self.max_shift

    #     return np.asarray([self.decode(pattern[..., i]) for i in range(pattern.shape[-1])])



from ntn_torch.encodings import Thermometer

from stable_baselines3.common.torch_layers import BaseFeaturesExtractor
from stable_baselines3.common.preprocessing import get_flattened_obs_dim


class ThermometerExtractor(BaseFeaturesExtractor):
  # I'll for now expect the minimum and maximum of the thermometer to be passed as arguments, but if a use a 
  # env wrapper that normalizes observations, I could have a fixed, or at least default thermo interval
  def __init__(self, observation_space, low, high, res):
    super(ThermometerExtractor, self).__init__(observation_space, res*get_flattened_obs_dim(observation_space))

    self.thermometer = Thermometer(low, high, res)

  def forward(self, observations):
    return self.thermometer.encode(observations).reshape((observations.shape[0], -1))


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ntn_torch',
 'ntn_torch.rl',
 'ntn_torch.rl.common',
 'ntn_torch.rl.ntn_reinforce']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.21.4,<2.0.0',
 'stable-baselines3[extra]>=1.3.0,<2.0.0',
 'torch>=1.10.0,<2.0.0']

setup_kwargs = {
    'name': 'ntn-torch',
    'version': '0.2.2',
    'description': 'ntn_torch is a library to build n-tuple neural network models using the framework of pytorch',
    'long_description': None,
    'author': 'Rafael F. Katopodis',
    'author_email': 'rafaelkatopodis@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<3.11',
}


setup(**setup_kwargs)
